#include "core.h"

//    This file is part of JustAudio.

//    JustAudio is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustAudio is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with BashWire under the GPL.txt file. If not, see
//    <http://www.gnu.org/licenses/>.

Core::Core(QObject *parent, QStringList *filesup) : QObject(parent)
{
    // the Core class acts as a non-gui centralized class that house the majority
    // of all functions the application needs to operate.

    exts = filesup;
}

Core::~Core()
{
    // this simply writes the current state of volume, the dir check box and mute
    // options to persistant storage so the application can reload the same state
    // on startup.

    emit stop();

    QFile file(CONFIG_FILE, this);

    if (file.open(QFile::WriteOnly | QFile::Truncate))
    {
        // CONFIG_FILE format:
        // [volume],[play_dir_check],[mute]

        // data formats:
        // VOLUME         = numeric string
        // PLAY_DIR_CHECK = string ("Y" or "N")
        // MUTE           = string ("Y" or "N")

        file.write(QByteArray::number(volumeSlider->value()));
        file.write(",");

        if (playDirAction->isChecked()) file.write("Y");
        else                            file.write("N");

        file.write(",");

        if (muteAction->isChecked()) file.write("Y");
        else                         file.write("N");
    }

    file.close();
}

void Core::setFileWatcher(QFileSystemWatcher *watcher)
{
    connect(watcher, SIGNAL(fileChanged(QString)), this, SLOT(processfileChanged(QString)));
}

void Core::setPlayer(QMediaPlayer *player)
{
    // the Core class does not run funtions directly on the QMediaPlayer
    // class that lives in the main function so everything it needs to
    // do are connected via slots and signals.

    mplayer = player;

    connect(this, &Core::start, mplayer, &QMediaPlayer::play);
    connect(this, &Core::pause, mplayer, &QMediaPlayer::pause);
    connect(this, &Core::stop, mplayer, &QMediaPlayer::stop);
    connect(this, &Core::setFile, mplayer, &QMediaPlayer::setSource);
    connect(this, &Core::setSeek, mplayer, &QMediaPlayer::setPosition);

    connect(mplayer, &QMediaPlayer::playbackStateChanged, this, &Core::stateChanged);
    connect(mplayer, &QMediaPlayer::mediaStatusChanged, this, &Core::statusChanged);
    connect(mplayer, &QMediaPlayer::positionChanged, this, &Core::setPosition);
}

void Core::setAudioOutput(QAudioOutput *audoOut)
{
    // same as setPlayer() except this connects all slots/signals
    // specific to QAudioOutput.

    if (audoOut != nullptr)
    {
        blockSignals(false);

        connect(this, &Core::setVol, audoOut, &QAudioOutput::setVolume);
        connect(this, &Core::muteState, audoOut, &QAudioOutput::setMuted);
    }
    else
    {
        // audo playback won't be possible without a valid
        // AudioOutput object so all core signals will be blocked
        // to prevent undefined behaviour.

        blockSignals(true);
    }
}

void Core::updateSeekDisp()
{
    seekDisp->setText(QString::number(mplayer->position()) + "/" + QString::number(mplayer->duration()));
}

void Core::setGuiActions(QAction *playAct, QAction *pauseAct, QAction *stopAct, QAction *nextAct, QAction *prevAct, QAction *muteAct, QAction *dir)
{
    // the core class will not interact directly with the GUI, instead all actions are
    // abstracted via various QAction classes passed onto it from the main function.

    playAction    = playAct;
    pauseAction   = pauseAct;
    stopAction    = stopAct;
    nextAction    = nextAct;
    prevAction    = prevAct;
    muteAction    = muteAct;
    playDirAction = dir;
    tracking      = true;

    connect(playAction, &QAction::triggered, this, &Core::start);
    connect(pauseAction, &QAction::triggered, this, &Core::pause);
    connect(stopAction, &QAction::triggered, this, &Core::stop);
    connect(nextAction, &QAction::triggered, this, &Core::nextFile);
    connect(prevAction, &QAction::triggered, this, &Core::prevFile);
    connect(muteAction, &QAction::triggered, this, &Core::muteState);

    playAction->setEnabled(false);  // play, pause, stop, previous and next are disabled initially.
    pauseAction->setEnabled(false); // this will change when an audio file is opened
    stopAction->setEnabled(false);  // (see stateChanged())
    nextAction->setEnabled(false);
    prevAction->setEnabled(false);
}

void Core::setGuiElements(QSlider *volSlider, QSlider *skSlider, QLabel *nameDisp, QLabel *seekLabel)
{
    volumeSlider = volSlider;
    seekSlider   = skSlider;
    currentFile  = nameDisp;
    seekDisp     = seekLabel;

    connect(seekSlider, &QSlider::sliderPressed, this, &Core::seekerPress);
    connect(seekSlider, &QSlider::sliderReleased, this, &Core::seekerRelease);
    connect(volumeSlider, &QSlider::valueChanged, this, &Core::setVolume);
    connect(mplayer, &QMediaPlayer::durationChanged, this, &Core::setDuration);

    // QMediaPlayer volume range 0-100.

    seekSlider->setMinimum(0);
    seekSlider->setMaximum(0);
    volumeSlider->setMinimum(0);
    volumeSlider->setMaximum(100);
    currentFile->setWordWrap(true);

    QFile file(CONFIG_FILE, this);

    bool ok = false;

    if (file.open(QFile::ReadOnly))
    {
        // the CONFIG_FILE is formated in csv so all data is seperated by a ','
        // the data will simply not load if the strings are not in exactly the
        // same order as it was saved.
        // data formats:
        // VOLUME         = numeric string
        // PLAY_DIR_CHECK = string ("Y" or "N")
        // MUTE           = string ("Y" or "N")

        auto confData = file.readAll();
        auto params   = confData.split(',');

        if (params.size() == 3)
        {
            int volume = params[VOLUME].toInt(&ok);

            if (ok)
            {
                volumeSlider->setValue(volume);

                setVolume(volume);

                if (params[PLAY_DIR_CHECK] == "Y") emit playDirAction->triggered(true);
                else                               emit playDirAction->triggered(false);

                if (params[MUTE] == "Y") emit muteAction->triggered(true);
                else                     emit muteAction->triggered(false);
            }
        }
    }

    if (!ok)
    {
        // set default values on failure to read the CONFIG_FILE.

        setVolume(50);

        emit playDirAction->triggered(false);
        emit muteAction->triggered(false);

        volumeSlider->setValue(50);
    }

    file.close();
}

void Core::processfileChanged(const QString &path)
{
    Q_UNUSED(path);

    QFile file(PROCESS_FILE, this);

    if (file.open(QFile::ReadOnly))
    {
        // expected data from the PROCESS_FILE should just be a directory path
        // string to the audio file to be opened. play() will call out an error
        // if it fails to open this file.

        play(file.readAll().trimmed());
    }

    file.close();
}

void Core::seekerPress()
{
    tracking = false;
}

void Core::seekerRelease()
{
    emit setSeek(seekSlider->value());

    tracking = true;
}

void Core::setDuration(int dur)
{
    seekSlider->setMaximum(dur);

    updateSeekDisp();
}

void Core::setPosition(int pos)
{
    if (tracking) seekSlider->setValue(pos);

    updateSeekDisp();
}

void Core::setVolume(int intValue)
{
    emit setVol(static_cast<float>(intValue) / static_cast<float>(volumeSlider->maximum()) * static_cast<float>(1));
}

void Core::play(const QString &path)
{
    emit stop();

    currentFile->setText(QFileInfo(path).fileName());

    emit setFile(QUrl::fromLocalFile(path));
    emit start();
}

void Core::nextFile()
{
    fileDir('+');
}

void Core::prevFile()
{
    fileDir('-');
}

void Core::fileDir(char direction)
{
    // using the current file as a reference, this function will first find out the
    // directory the file lives in and then use that directory to list all files within
    // it. then using the current file as a reference again, it trys to find it in the
    // list of files and then use that file position to find out the next or previous
    // file in the directory to play.

    // if the reference file no longer exist then it will attempt to play 1st file in the
    // directory. if the directory no longer exist then nothing happens.

    QFileInfo info(mplayer->source().path());
    QDir      dir(info.path());

    auto list = dir.entryList(*exts, QDir::Files | QDir::Readable, QDir::Name | QDir::IgnoreCase);
    auto pos  = list.indexOf(info.fileName());

    if (!list.isEmpty())
    {
        if      (direction == '+') pos++;
        else if (direction == '-') pos--;

        if (pos < 0) pos = 0;

        if (pos < list.size())
        {
            play(info.path() + QDir::separator() + list[pos]);
        }
    }
}

void Core::stateChanged(QMediaPlayer::PlaybackState state)
{
    switch(state)
    {
    case QMediaPlayer::PlayingState:
    {
        playAction->setEnabled(false);
        pauseAction->setEnabled(true);
        stopAction->setEnabled(true);
        prevAction->setEnabled(true);
        nextAction->setEnabled(true);

        break;
    }
    case QMediaPlayer::StoppedState:
    {
        playAction->setEnabled(true);
        pauseAction->setEnabled(false);
        stopAction->setEnabled(false);

        break;
    }
    case QMediaPlayer::PausedState:
    {
        playAction->setEnabled(true);
        pauseAction->setEnabled(false);
        stopAction->setEnabled(true);

        break;
    }
    }
}

void Core::statusChanged(QMediaPlayer::MediaStatus state)
{
    if (state == QMediaPlayer::EndOfMedia)
    {
        if (playDirAction->isChecked()) nextFile();
    }
}

void Core::volumeDown()
{
    volumeSlider->setValue(volumeSlider->value() - 5);

    setVolume(volumeSlider->value());
}

void Core::volumeUp()
{
    volumeSlider->setValue(volumeSlider->value() + 5);

    setVolume(volumeSlider->value());
}
